/* tile editor 4.0 */
#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <string.h>

#define ID_Exit		1
#define ID_About	2
#define ID_Selector	3
#define ID_Open		4
#define ID_Save		5
#define ID_Palette	6
#define ID_Minipal	7
#define ID_Prop		8
#define ID_Undo		24

char *tileset_name[] = {
	"0 - Underworld",
	"1 - Onett",
	"2 - Twoson",
	"3 - Threed",
	"4 - Fourside",
	"5 - Magicant",
	"6 - Outdoors",
	"7 - Summers",
	"8 - Desert",
	"9 - Dalaam",
	"10 - Indoors 1",
	"11 - Indoors 2",
	"12 - Stores 1",
	"13 - Caves 1",
	"14 - Indoors 3",
	"15 - Stores 2",
	"16 - Indoors 4",
	"17 - Winters",
	"18 - Scaraba",
	"19 - Caves 2"
};

struct {
	short tileset, tile, val;
	char action, x, y, ch;
} undo[100];
unsigned char tilesets[20][1024][32];
unsigned short arranges[20][1024][16];
unsigned char collision[20][1024][16];
unsigned char rom[4194816];
int rom_loaded, tileset, color, palette, minipal, tile, arrangement, arrangel;
char tileset_changed[20], arrange_changed[20];
int pal[6][16];
HWND mainwindow, tilewindow, edit1window, edit2window, palwindow, arrangewindow, aeditwindow, peditwindow, ceditwindow;
HWND rscroll, gscroll, bscroll;
struct {
	struct {
		char mtileset, mpalette;
		int start;
	} pal[60];
	int len;
} paldata[20];

#define tileset_pointers ((int *)&rom[0x2F125B])
#define arrange_pointers ((int *)&rom[0x2F12AB])
#define palette_pointers ((int *)&rom[0x2F12FB])
#define collision_pointers ((int *)&rom[0x2F137B])

void update(void) {
	int i;
	char buf[3];
	InvalidateRgn(tilewindow, NULL, 0);
	InvalidateRgn(edit1window, NULL, 0);
	InvalidateRgn(edit2window, NULL, 0);
	InvalidateRgn(palwindow, NULL, 0);
	InvalidateRgn(arrangewindow, NULL, 0);
	InvalidateRgn(aeditwindow, NULL, 0);
	for(i = 0; i < 16; i++) {
		sprintf(buf, "%02x", collision[tileset][arrangement][i]);
		SendDlgItemMessage(ceditwindow, ID_Prop + i, WM_SETTEXT, 0, (LPARAM)buf);
	}
}

void do_undo(void) {
	unsigned char *b;
	switch(undo->action) {
		case 0:
			MessageBox(mainwindow, "Nothing to undo", NULL, MB_OK);
			break;
		case 1:
			b = &tilesets[undo->tileset][undo->tile][2 * undo->y];
			b[ 0] &= ~(128 >> undo->x); if(undo->val & 1) b[ 0] |= (128 >> undo->x);
			b[ 1] &= ~(128 >> undo->x); if(undo->val & 2) b[ 1] |= (128 >> undo->x);
			b[16] &= ~(128 >> undo->x); if(undo->val & 4) b[16] |= (128 >> undo->x);
			b[17] &= ~(128 >> undo->x); if(undo->val & 8) b[17] |= (128 >> undo->x);
			tileset_changed[undo->tileset] = undo->ch;
			memmove(undo, &undo[1], sizeof(*undo) * 99);
			undo[99].action = 0;
			break;
		case 2:
			arranges[undo->tileset][undo->tile][4 * undo->y + undo->x] = undo->val;
			tileset_changed[undo->tileset] = undo->ch;
			memmove(undo, &undo[1], sizeof(*undo) * 99);
			undo[99].action = 0;
			break;
	}
	update();
}

void updatepal(void) {
	int i, x, y;
	for(i = 0; i < 6; i++) {
		for(x = 0; x < 16; x++) {
			y = *(unsigned short *)&rom[paldata[tileset].pal[palette].start + 32*i + 2*x];
			pal[i][x]  = (((y & 0x1F) >> 0) * 8) << 16;
			pal[i][x] |= (((y & 0x3E0) >> 5) * 8) << 8;
			pal[i][x] |= (((y & 0x7C00) >> 10) * 8) << 0;
		}
	}
	SetScrollPos(rscroll, SB_CTL, (pal[minipal][color] & 0xFF0000) >> 19, 1);
	SetScrollPos(gscroll, SB_CTL, (pal[minipal][color] & 0x00FF00) >> 11, 1);
	SetScrollPos(bscroll, SB_CTL, (pal[minipal][color] & 0x0000FF) >> 3, 1);
	update();
}

int tilecolor(unsigned char *tile, int x, int y) {
	int c = 0;
	if(tile[ 0 + 2*y] & (128 >> x)) c  = 1;
	if(tile[ 1 + 2*y] & (128 >> x)) c |= 2;
	if(tile[16 + 2*y] & (128 >> x)) c |= 4;
	if(tile[17 + 2*y] & (128 >> x)) c |= 8;
	return c;
}

/* Used internally by comp */
unsigned char *bpos, *pos;
void encode(int length, int type) {
	if(length > 32) {
		*bpos++ = 0xE0 + 4 * type + ((length - 1) >> 8);
		*bpos++ = (length - 1) & 0xFF;
	} else
		*bpos++ = 0x20 * type + length - 1;
}
void rencode(int length) {
	if(length <= 0) return;
	encode(length, 0);
	memcpy(bpos, pos, length);
	bpos += length;
}

/* The compressor function.
 * Takes a pointer to the uncompressed block, a pointer to 65536 bytes
 * which it compresses into, and the size of the uncompressed block.
 * Returns the number of bytes compressed.
 */
int comp(unsigned char *udata, unsigned char *buffer, int length) {
	unsigned char *limit = &udata[length];
	unsigned char *pos2, *pos3, bitrevs[256];
	int tmp;
	bpos = buffer; pos = udata;
	for(tmp = 0; tmp < 256; tmp++) {
		unsigned char x = tmp;
		x = ((x >> 1) & 0x55) | ((x << 1) & 0xAA);
		x = ((x >> 2) & 0x33) | ((x << 2) & 0xCC);
		bitrevs[tmp] = ((x >> 4) & 0x0F) | ((x << 4) & 0xF0);
	}
	while(pos < limit) {
		/* Look for patterns */
		for(pos2 = pos; pos2 < limit && pos2 < pos + 1024; pos2++) {
			for(pos3 = pos2; pos3 < limit && pos3 < pos2 + 1024 && *pos2 == *pos3; pos3++);
			if(pos3 - pos2 >= 3) {
				rencode(pos2 - pos);
				encode(pos3 - pos2, 1);
				*bpos++ = *pos2;
				pos = pos3;
				break;
			}
			for(pos3 = pos2; pos3 < limit && pos3 < pos2 + 2048 && *pos3 == *pos2 && pos3[1] == pos2[1]; pos3 += 2);
			if(pos3 - pos2 >= 6) {
				rencode(pos2 - pos);
				encode((pos3 - pos2) / 2, 2);
				*bpos++ = pos2[0];
				*bpos++ = pos2[1];
				pos = pos3;
				break;
			}
			for(tmp = 0, pos3 = pos2; pos3 < limit && pos3 < pos2 + 1024 && *pos3 == *pos2 + tmp; pos3++, tmp++);
			if(pos3 - pos2 >= 4) {
				rencode(pos2 - pos);
				encode(pos3 - pos2, 3);
				*bpos++ = *pos2;
				pos = pos3;
				break;
			}
			for(pos3 = udata; (pos3 = memchr(pos3, *pos2, pos2 - pos3)); pos3++) {
				for(tmp = 0; pos3[tmp] == pos2[tmp] && tmp < pos2 - pos3 && tmp < 1024; tmp++);
				if(tmp >= 5) {
					rencode(pos2 - pos);
					encode(tmp, 4);
					*bpos++ = (pos3 - udata) >> 8;
					*bpos++ = (pos3 - udata) & 0xFF;
					pos = pos2 + tmp;
					goto DONE;
				}
				for(tmp = 0; pos3[-tmp] == pos2[tmp] && tmp <= pos3 - udata && tmp < 1024; tmp++);
				if(tmp >= 5) {
					rencode(pos2 - pos);
					encode(tmp, 6);
					*bpos++ = (pos3 - udata) >> 8;
					*bpos++ = (pos3 - udata) & 0xFF;
					pos = pos2 + tmp;
					goto DONE;
				}
			}
			for(pos3 = udata; (pos3 = memchr(pos3, bitrevs[*pos2], pos2 - pos3)); pos3++) {
				for(tmp = 0; pos3[tmp] == bitrevs[pos2[tmp]] && tmp < pos2 - pos3 && tmp < 1024; tmp++);
				if(tmp >= 5) {
					rencode(pos2 - pos);
					encode(tmp, 5);
					*bpos++ = (pos3 - udata) >> 8;
					*bpos++ = (pos3 - udata) & 0xFF;
					pos = pos2 + tmp;
					goto DONE;
				}
			}
		}
		DONE:
		/* Can't compress, so just use 0 (raw) */
		rencode(pos2 - pos);
		if(pos < pos2) pos = pos2;
	}
	*bpos++ = 0xFF;
	return bpos - buffer;
}

/* The decompressor function.
 * Takes a pointer to the compressed block, a pointer to the buffer
 * which it decompresses into, and the maximum length. Returns the
 * number of bytes uncompressed, or -1 if decompression failed.
 */
int decomp(unsigned char *cdata, unsigned char *buffer, int maxlen) {
	unsigned char *bpos = buffer, *bpos2, tmp;
	while(*cdata != 0xFF) {
		int cmdtype = *cdata >> 5;
		int len = (*cdata & 0x1F) + 1;
		if(cmdtype == 7) {
			cmdtype = (*cdata & 0x1C) >> 2;
			len = ((*cdata & 3) << 8) + *(cdata + 1) + 1;
			cdata++;
		}
		if(bpos + len > &buffer[maxlen]) return -1;
		cdata++;
		if(cmdtype >= 4) {
			bpos2 = &buffer[(*cdata << 8) + *(cdata + 1)];
			if(bpos2 >= &buffer[maxlen]) return -1;
			cdata += 2;
		}
		switch(cmdtype) {
			case 0:
				memcpy(bpos, cdata, len);
				cdata += len;
				bpos += len;
				break;
			case 1:
				memset(bpos, *cdata++, len);
				bpos += len;
				break;
			case 2:
				if(bpos + 2 * len > &buffer[maxlen]) return -1;
				while(len--) {
					*(short *)bpos = *(short *)cdata;
					bpos += 2;
				}
				cdata += 2;
				break;
			case 3:
				tmp = *cdata++;
				while(len--) *bpos++ = tmp++;
				break;
			case 4:
				if(bpos2 + len > &buffer[maxlen]) return -1;
				memcpy(bpos, bpos2, len);
				bpos += len;
				break;
			case 5:
				if(bpos2 + len > &buffer[maxlen]) return -1;
				while(len--) {
					tmp = *bpos2++;
					/* reverse the bits */
					tmp = ((tmp >> 1) & 0x55) | ((tmp << 1) & 0xAA);
					tmp = ((tmp >> 2) & 0x33) | ((tmp << 2) & 0xCC);
					tmp = ((tmp >> 4) & 0x0F) | ((tmp << 4) & 0xF0);
					*bpos++ = tmp;
				}
				break;
			case 6:
				if(bpos2 - len + 1 < buffer) return -1;
				while(len--) *bpos++ = *bpos2--;
				break;
			case 7:
				return -1;
		}
	}
	return bpos - buffer;
}

/* window event handler */
LRESULT CALLBACK main_WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
	if(msg == WM_SYSCOMMAND && wParam == SC_HOTKEY) do_undo();
	switch(msg) {
		case WM_COMMAND: switch(LOWORD(wParam)) {
			case ID_Exit:
				PostMessage(hWnd, WM_CLOSE, 0, 0);
				break;
			case ID_Open: {
				OPENFILENAME ofn;
				FILE *f;
				char filename[MAX_PATH] = "";
				int i, j, k;
				ZeroMemory(&ofn, sizeof(ofn));
				ofn.lStructSize = sizeof(ofn);
				ofn.hwndOwner = hWnd;
				ofn.lpstrFilter = "SNES Rom files (*.smc)\0*.smc\0All Files (*.*)\0*.*\0";
				ofn.lpstrFile = filename;
				ofn.nMaxFile = MAX_PATH;
				ofn.Flags = OFN_EXPLORER | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
				ofn.lpstrDefExt = "smc";
				if(GetOpenFileName(&ofn)) {
					f = fopen(filename, "rb");
					if(f) {
						if(fread(rom, 1, 0x400200, f) == 0x300200)
							for(i = 0; i < 4096; i++)
								rom[0x3002FF + 256*i] = 0x02;
						fclose(f);
						rom_loaded = 1;
						for(i = 0; i < 20; i++) {
							if(decomp(&rom[tileset_pointers[i] - 0xBFFE00], (char *)tilesets[i], 28673) != 28673 || decomp(&rom[arrange_pointers[i] - 0xBFFE00], (char *)arranges[i], 32768) == -1) { 
								MessageBox(hWnd, "Bad compressed data", NULL, MB_OK);
								return 0;
							}
							tileset_changed[i] = tileset_pointers[i] >= 0xF00000;
							arrange_changed[i] = arrange_pointers[i] >= 0xF00000;
							for(j = 0; j < 1024; j++)
								for(k = 0; k < 16; k++)
									collision[i][j][k] = rom[((unsigned short *)&rom[collision_pointers[i] - 0xBFFE00])[j] + 0x180200 + k];
							paldata[i].len = 0;
						}
						for(i = 0; i < 32; i++) {
							int t = rom[0x2F121B + 2 * i];
							k = palette_pointers[i + 1] - palette_pointers[i];
							if(i == 31) k = 0xDAFAA7 - palette_pointers[i];
							for(j = 0; j < k / 0xC0; j++) {
								paldata[t].pal[paldata[t].len].mtileset = i;
								paldata[t].pal[paldata[t].len].mpalette = j;
								paldata[t].pal[paldata[t].len].start = palette_pointers[i] - 0xBFFE00 + 0xC0 * j;
								paldata[t].len++;
							}
						}
						goto newtileset;
					} else {
						MessageBox(hWnd, "rb", NULL, MB_OK);
					}
				}
			} break;
			case ID_Save: {
				OPENFILENAME ofn;
				FILE *f;
				char filename[MAX_PATH] = "";
				int i, j, k, l;
				int cpos = 0x3A0200;
				ZeroMemory(&ofn, sizeof(ofn));
				ofn.lStructSize = sizeof(ofn);
				ofn.hwndOwner = hWnd;
				ofn.lpstrFilter = "SNES Rom files (*.smc)\0*.smc\0All Files (*.*)\0*.*\0";
				ofn.lpstrFile = filename;
				ofn.nMaxFile = MAX_PATH;
				ofn.Flags = OFN_EXPLORER | OFN_HIDEREADONLY;
				ofn.lpstrDefExt = "smc";
				if(GetSaveFileName(&ofn)) {
					f = fopen(filename, "wb");
					if(f) {
						char buf[30];
						int cc = 0;
						short cp[20480];
						int cpi = 0;
						for(i = 0; i < 20; i++) {
							if(tileset_changed[i]) {
								tileset_pointers[i] = cpos + 0xBFFE00;
								cpos += comp((char *)tilesets[i], &rom[cpos], 28673);  
							}
							if(arrange_changed[i]) {
								arrange_pointers[i] = cpos + 0xBFFE00;
								for(j = 1024; j; j--)
									for(k = 0; k < 16; k++)
										if(arranges[i][j - 1][k]) goto A;
								A: cpos += comp((char *)arranges[i], &rom[cpos], j * 32);
							}
							for(j = 1024; j; j--)
								for(k = 0; k < 16; k++)
									if(arranges[i][j - 1][k]) goto B;
							B:
							collision_pointers[i] = cpi * 2;
							for(k = 0; k < j; k++) {
								for(l = 0; l < cc; l++)
									if(!memcmp(&rom[0x180200 + 16 * l], collision[i][k], 16)) goto C;
								memcpy(&rom[0x180200 + 16 * cc++], collision[i][k], 16);
								C: cp[cpi++] = l * 16;
							}
						}
						memcpy(&rom[0x180200 + 16 * cc], cp, 2 * cpi);
						for(i = 0; i < 20; i++)
							collision_pointers[i] = collision_pointers[i] + 0xD80000 + 16 * cc;
						fwrite(rom, 4194816, 1, f);
						fclose(f);
						sprintf(buf, "Stored from 3A0200 to %06X", cpos - 1);
						MessageBox(hWnd, buf, "Saved", MB_OK);
					} else {
						MessageBox(hWnd, "wb", NULL, MB_OK);
					}
				}
			} break;
			case ID_About:
				MessageBox(hWnd, 
					"EarthBound Tile Editor version 4.0\n"
					"Created by cabbage, a.k.a. mrnobo1024\n"
					"Visit http://pkhack.starmen.net/ for more EB hacking programs and info",
					"About", MB_OK);
				break;
			case ID_Undo: do_undo(); break;
			case ID_Selector: if(HIWORD(wParam) == CBN_SELCHANGE) {
				int i;
				char buf[5];
				tileset = SendDlgItemMessage(hWnd,
					ID_Selector, CB_GETCURSEL, 0, 0);
				newtileset:
				SendDlgItemMessage(hWnd, ID_Palette, CB_RESETCONTENT, 0, 0);
				for(i = 0; i < paldata[tileset].len; i++) {
					sprintf(buf, "%d/%d", paldata[tileset].pal[i].mtileset, paldata[tileset].pal[i].mpalette);
					SendDlgItemMessage(hWnd, ID_Palette, CB_ADDSTRING, 0, (LPARAM)buf);
				}
				sprintf(buf, "%d/%d", paldata[tileset].pal[0].mtileset, paldata[tileset].pal[0].mpalette);
				SendDlgItemMessage(hWnd, ID_Palette, CB_SELECTSTRING, 0, (LPARAM)buf);
				palette = 0;
				updatepal();
			} break;
			case ID_Palette: if(HIWORD(wParam) == CBN_SELCHANGE) {
				palette = SendDlgItemMessage(hWnd, ID_Palette, CB_GETCURSEL, 0, 0);
				updatepal();

			} break;
			case ID_Minipal: if(HIWORD(wParam) == CBN_SELCHANGE) {
				minipal = SendDlgItemMessage(hWnd, ID_Minipal, CB_GETCURSEL, 0, 0);
				updatepal();
			} break;
		} break;
		case WM_CLOSE: DestroyWindow(hWnd); break;
		case WM_DESTROY: PostQuitMessage(0); break;
		case WM_PAINT: {
			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(hWnd, &ps);
			TextOut(hdc, 520, 3, "Tileset", 7);
			TextOut(hdc, 520, 50, "Palette", 7);
			TextOut(hdc, 520, 80, "Subpalette", 10);
			EndPaint(hWnd, &ps);
		} break;
		default: return DefWindowProc(hWnd, msg, wParam, lParam);
	}
	return 0;
}

LRESULT CALLBACK tiles_WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
		case WM_LBUTTONDOWN: {
			int x = LOWORD(lParam);
			int y = HIWORD(lParam);
			tile = (y / 16) * 32 + (x / 16);
			update();
		} break;
		case WM_PAINT: {
			BITMAPINFO bmi = { { 40, 512, -256, 1, 32, BI_RGB, 0, 0, 0, 0, 0 } };
			HBITMAP bm;
			int *data;
			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(hWnd, &ps);
			int x, y, tx, ty;
			bm = CreateDIBSection(hdc, &bmi, DIB_RGB_COLORS, (void **)&data, NULL, 0);
			for(y = 0; y < 16; y++) for(x = 0; x < 32; x++) {
				for(ty = 0; ty < 8; ty++) for(tx = 0; tx < 8; tx++) {
					data[16*x + 2*tx     + 512 * (16*y + 2*ty)] =
					data[16*x + 2*tx + 1 + 512 * (16*y + 2*ty)] =
					data[16*x + 2*tx     + 512 * (16*y + 2*ty + 1)] =
					data[16*x + 2*tx + 1 + 512 * (16*y + 2*ty + 1)] =
						pal[minipal][tilecolor(tilesets[tileset][y * 32 + x], tx, ty)];
				}
			}
			SetDIBitsToDevice(hdc, 0, 0, 512, 256, 0, 0, 0, 256, data, &bmi, DIB_RGB_COLORS);
			DeleteObject(bm);
			EndPaint(hWnd, &ps);
		} break;
		default: return DefWindowProc(hWnd, msg, wParam, lParam);
	}
	return 0;
}

LRESULT CALLBACK edit1_WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
		case WM_LBUTTONDOWN: case WM_MOUSEMOVE: {
			int x = LOWORD(lParam);
			int y = HIWORD(lParam);
			unsigned char *b;
			if(!(wParam & MK_LBUTTON) || !rom_loaded) break;
			x /= 8; y /= 8;
			b = &tilesets[tileset][tile][y * 2];
			if(tilecolor(tilesets[tileset][tile], x, y) == color) break;
			memmove(&undo[1], undo, sizeof(*undo) * 99);
			undo->tileset = tileset;
			undo->action = 1;
			undo->tile = tile;
			undo->x = x;
			undo->y = y;
			undo->val = tilecolor(tilesets[tileset][tile], x, y);
			undo->ch = tileset_changed[tileset];
			b[ 0] &= ~(128 >> x); if(color & 1) b[ 0] |= (128 >> x);
			b[ 1] &= ~(128 >> x); if(color & 2) b[ 1] |= (128 >> x);
			b[16] &= ~(128 >> x); if(color & 4) b[16] |= (128 >> x);
			b[17] &= ~(128 >> x); if(color & 8) b[17] |= (128 >> x);
			tileset_changed[tileset] = 1;
			update();
		} break;
		case WM_PAINT: {
			BITMAPINFO bmi = { { 40, 64, -64, 1, 32, BI_RGB, 0, 0, 0, 0, 0 } };
			HBITMAP bm;
			int *data;
			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(hWnd, &ps);
			int x, y, tx, ty;
			bm = CreateDIBSection(hdc, &bmi, DIB_RGB_COLORS, (void **)&data, NULL, 0);
			for(ty = 0; ty < 8; ty++) for(tx = 0; tx < 8; tx++)
				for(y = 0; y < 8; y++) for(x = 0; x < 8; x++) {
					data[8*tx + x + 64 * (8*ty + y)] = pal[minipal][tilecolor(tilesets[tileset][tile], tx, ty)];
				}
			SetDIBitsToDevice(hdc, 0, 0, 64, 64, 0, 0, 0, 64, data, &bmi, DIB_RGB_COLORS);
			DeleteObject(bm);
			EndPaint(hWnd, &ps);
		} break;
		default: return DefWindowProc(hWnd, msg, wParam, lParam);
	}
	return 0;
}

LRESULT CALLBACK edit2_WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
		case WM_LBUTTONDOWN: case WM_MOUSEMOVE: {
			int x = LOWORD(lParam);
			int y = HIWORD(lParam);
			unsigned char *b;
			if(!(wParam & MK_LBUTTON) || !rom_loaded) break;
			x /= 8; y /= 8;
			b = &tilesets[tileset][tile + 512][y * 2];
			if(tilecolor(tilesets[tileset][tile + 512], x, y) == color) break;
			memmove(&undo[1], undo, sizeof(*undo) * 99);
			undo->tileset = tileset;
			undo->action = 1;
			undo->tile = tile + 512;
			undo->x = x;
			undo->y = y;
			undo->val = tilecolor(tilesets[tileset][tile + 512], x, y);
			undo->ch = tileset_changed[tileset];
			b[ 0] &= ~(128 >> x); if(color & 1) b[ 0] |= (128 >> x);
			b[ 1] &= ~(128 >> x); if(color & 2) b[ 1] |= (128 >> x);
			b[16] &= ~(128 >> x); if(color & 4) b[16] |= (128 >> x);
			b[17] &= ~(128 >> x); if(color & 8) b[17] |= (128 >> x);
			tileset_changed[tileset] = 1;
			update();
		} break;
		case WM_PAINT: {
			BITMAPINFO bmi = { { 40, 64, -64, 1, 32, BI_RGB, 0, 0, 0, 0, 0 } };
			HBITMAP bm;
			int *data;
			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(hWnd, &ps);
			int x, y, tx, ty;
			bm = CreateDIBSection(hdc, &bmi, DIB_RGB_COLORS, (void **)&data, NULL, 0);
			for(ty = 0; ty < 8; ty++) for(tx = 0; tx < 8; tx++)
				for(y = 0; y < 8; y++) for(x = 0; x < 8; x++) {
					data[8*tx + x + 64 * (8*ty + y)] = pal[minipal][tilecolor(tilesets[tileset][tile + 512], tx, ty)];
				}
			SetDIBitsToDevice(hdc, 0, 0, 64, 64, 0, 0, 0, 64, data, &bmi, DIB_RGB_COLORS);
			DeleteObject(bm);
			EndPaint(hWnd, &ps);
		} break;
		default: return DefWindowProc(hWnd, msg, wParam, lParam);
	}
	return 0;
}	

LRESULT CALLBACK pal_WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
		case WM_LBUTTONDOWN: {
			int x = LOWORD(lParam);
			int y = HIWORD(lParam);
			color = (y / 16) * 8 + (x / 16);
			updatepal();
		} break;
		case WM_PAINT: {
			BITMAPINFO bmi = { { 40, 128, -32, 1, 32, BI_RGB, 0, 0, 0, 0, 0 } };
			HBITMAP bm;
			int *data;
			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(hWnd, &ps);
			int x, y, tx, ty;
			bm = CreateDIBSection(hdc, &bmi, DIB_RGB_COLORS, (void **)&data, NULL, 0);
			for(y = 0; y < 2; y++) for(x = 0; x < 8; x++)
				for(ty = 0; ty < 16; ty++) for(tx = 0; tx < 16; tx++)
					data[16*x + tx + 128 * (16*y + ty)] = pal[minipal][y * 8 + x];
			SetDIBitsToDevice(hdc, 0, 0, 128, 32, 0, 0, 0, 32, data, &bmi, DIB_RGB_COLORS);
			DeleteObject(bm);
			EndPaint(hWnd, &ps);
		} break;
		default: return DefWindowProc(hWnd, msg, wParam, lParam);
	}
	return 0;
}

LRESULT CALLBACK arrange_WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
		case WM_HSCROLL: switch(LOWORD(wParam)) {
			case SB_LINELEFT: case SB_PAGELEFT: if(arrangel) arrangel--; break;
			case SB_LINERIGHT: case SB_PAGERIGHT: if(arrangel < 240) arrangel++; break;
			case SB_THUMBTRACK: arrangel = HIWORD(wParam);
		}
		SetScrollPos(hWnd, SB_HORZ, arrangel, 1);
		InvalidateRgn(arrangewindow, NULL, 0);
		break;
		case WM_LBUTTONDOWN: {
			int x = LOWORD(lParam);
			int y = HIWORD(lParam);
			arrangement = (x / 33 + arrangel) * 4 + y / 33;
			update();
		} break;
		case WM_PAINT: {
			BITMAPINFO bmi = { { 40, 527, -131, 1, 32, BI_RGB, 0, 0, 0, 0, 0 } };
			HBITMAP bm;
			int *data;
			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(hWnd, &ps);
			int xx, yy, x, y, tx, ty;
			bm = CreateDIBSection(hdc, &bmi, DIB_RGB_COLORS, (void **)&data, NULL, 0);
			for(yy = 0; yy < 4; yy++) for(xx = 0; xx < 15; xx++) 
				for(y = 0; y < 4; y++) for(x = 0; x < 4; x++) {
					int b = arranges[tileset][(xx + arrangel) * 4 + yy][y * 4 + x];
					for(ty = 0; ty < 8; ty++) for(tx = 0; tx < 8; tx++) {
						data[33*xx + 8*x + (b & 0x4000 ? 7 - tx : tx) + 527 * 
							(33*yy + 8*y + (b & 0x8000 ? 7 - ty : ty))] =
							pal[((b & 0x1C00) >> 10) - 2][
							tilecolor(tilesets[tileset][b & 0x1FF], tx, ty)];
					}
				}
			SetDIBitsToDevice(hdc, 0, 0, 527, 131, 0, 0, 0, 131, data, &bmi, DIB_RGB_COLORS);
			DeleteObject(bm);
			EndPaint(hWnd, &ps);
		} break;
		default: return DefWindowProc(hWnd, msg, wParam, lParam);
	}
	return 0;
}

LRESULT CALLBACK pedit_WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
		case WM_HSCROLL: {
			int p = GetScrollPos((HWND)lParam, SB_CTL);
			int r, g, b;
			switch(LOWORD(wParam)) {
				case SB_LINELEFT: case SB_PAGELEFT: if(p > 0) p--; break;
				case SB_LINERIGHT: case SB_PAGERIGHT: if(p < 31) p++; break;
				case SB_THUMBTRACK: p = HIWORD(wParam); break;
			}
			SetScrollPos((HWND)lParam, SB_CTL, p, 1);
			r = GetScrollPos(rscroll, SB_CTL);
			g = GetScrollPos(gscroll, SB_CTL);
			b = GetScrollPos(bscroll, SB_CTL);
			*(unsigned short *)&rom[paldata[tileset].pal[palette].start + 32*minipal + 2*color] = b << 10 | g << 5 | r;
			updatepal();
		} break;
		case WM_PAINT: {
			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(hWnd, &ps);
			char buf[5];
			sprintf(buf, "R%-3d", GetScrollPos(rscroll, SB_CTL));
			TextOut(hdc, 100,  0, buf, 4);
			sprintf(buf, "G%-3d", GetScrollPos(gscroll, SB_CTL));
			TextOut(hdc, 100, 16, buf, 4);
			sprintf(buf, "B%-3d", GetScrollPos(bscroll, SB_CTL));
			TextOut(hdc, 100, 32, buf, 4);
			EndPaint(hWnd, &ps);
		} break;
		default: return DefWindowProc(hWnd, msg, wParam, lParam);
	}
	return 0;
}

LRESULT CALLBACK aedit_WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
		case WM_LBUTTONDOWN: {
			int x = LOWORD(lParam) / 17;
			int y = HIWORD(lParam) / 17;
			memmove(&undo[1], undo, sizeof(*undo) * 99);
			if(wParam & MK_SHIFT) {
				tile = arranges[tileset][arrangement][y * 4 + x] & 0x1FF;
				update();
			} else {
				undo->tileset = tileset;
				undo->action = 2;
				undo->tile = arrangement;
				undo->x = x;
				undo->y = y;
				undo->val = arranges[tileset][arrangement][y * 4 + x];
				undo->ch = arrange_changed[tileset];
				arranges[tileset][arrangement][y * 4 + x] = tile + 0x800 + minipal * 0x400;
				update();
				arrange_changed[tileset] = 1;
			}
		} break;
		case WM_RBUTTONDOWN: {
			int x = LOWORD(lParam) / 17;
			int y = HIWORD(lParam) / 17;
			memmove(&undo[1], undo, sizeof(*undo) * 99);
			undo->tileset = tileset;
			undo->action = 2;
			undo->tile = arrangement;
			undo->x = x;
			undo->y = y;
			undo->val = arranges[tileset][arrangement][y * 4 + x];
			undo->ch = arrange_changed[tileset];
			arranges[tileset][arrangement][y * 4 + x] += 0x4000;
			update();
			arrange_changed[tileset] = 1;
		} break;
		case WM_PAINT: {
			BITMAPINFO bmi = { { 40, 67, -67, 1, 32, BI_RGB, 0, 0, 0, 0, 0 } };
			HBITMAP bm;
			int *data;
			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(hWnd, &ps);
			int x, y, tx, ty, xx, yy;
			bm = CreateDIBSection(hdc, &bmi, DIB_RGB_COLORS, (void **)&data, NULL, 0);
			for(y = 0; y < 4; y++) for(x = 0; x < 4; x++) {
				int b = arranges[tileset][arrangement][y * 4 + x];
				for(ty = 0; ty < 8; ty++) for(tx = 0; tx < 8; tx++) {
					for(yy = 0; yy < 2; yy++) for(xx = 0; xx < 2; xx++) {
						data[17*x + 2*(b & 0x4000 ? 7 - tx : tx) + xx + 67 * 
							(17*y + 2*(b & 0x8000 ? 7 - ty : ty) + yy)] =
							pal[((b & 0x1C00) >> 10) - 2][
							tilecolor(tilesets[tileset][b & 0x1FF], tx, ty)];
					}
				}
			}
			SetDIBitsToDevice(hdc, 0, 0, 67, 67, 0, 0, 0, 67, data, &bmi, DIB_RGB_COLORS);
			DeleteObject(bm);
			EndPaint(hWnd, &ps);
		} break;
		default: return DefWindowProc(hWnd, msg, wParam, lParam);
	}
	return 0;
}

LRESULT CALLBACK cedit_WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
		case WM_PARENTNOTIFY: {
			int x = LOWORD(lParam);
			int y = HIWORD(lParam);
			SetFocus(GetDlgItem(hWnd, ID_Prop + (y / 20) * 4 + (x / 20)));
		} break;
		case WM_CTLCOLOREDIT: {
			int i;
			char buf[10];
			*(int *)buf = 10;
			for(i = 0; i < 16; i++) {
				SendDlgItemMessage(hWnd, ID_Prop + i, EM_GETLINE, 0, (LPARAM)buf);
				collision[tileset][arrangement][i] = strtol(buf, NULL, 16);
			}
		} break;
		default: return DefWindowProc(hWnd, msg, wParam, lParam);
	}
	return 0;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
	LPSTR lpCmdLine, int nCmdShow)
{
	HWND listbox;
	MSG msg;
	int i;
	HMENU menu, submenu;

	int x = GetSystemMetrics(SM_CXFIXEDFRAME) * 2;
	int y = GetSystemMetrics(SM_CYFIXEDFRAME) * 2 + GetSystemMetrics(SM_CYCAPTION);

	RegisterClass(&(WNDCLASS){ 0, main_WndProc, 0, 0, hInstance, LoadIcon(NULL, IDI_APPLICATION), LoadCursor(NULL, IDC_ARROW), (HBRUSH)COLOR_WINDOWFRAME, NULL, "main" });
	RegisterClass(&(WNDCLASS){ 0, tiles_WndProc, 0, 0, hInstance, NULL, NULL, NULL, NULL, "tiles" });
	RegisterClass(&(WNDCLASS){ 0, edit1_WndProc, 0, 0, hInstance, NULL, NULL, NULL, NULL, "edit1" });
	RegisterClass(&(WNDCLASS){ 0, edit2_WndProc, 0, 0, hInstance, NULL, NULL, NULL, NULL, "edit2" });
	RegisterClass(&(WNDCLASS){ 0, pal_WndProc, 0, 0, hInstance, NULL, NULL, NULL, NULL, "pal" });
	RegisterClass(&(WNDCLASS){ 0, pedit_WndProc, 0, 0, hInstance, NULL, NULL, NULL, NULL, "pedit" });
	RegisterClass(&(WNDCLASS){ 0, arrange_WndProc, 0, 0, hInstance, NULL, NULL, NULL, NULL, "arrange" });
	RegisterClass(&(WNDCLASS){ 0, aedit_WndProc, 0, 0, hInstance, NULL, NULL, NULL, NULL, "aedit" });
	RegisterClass(&(WNDCLASS){ 0, cedit_WndProc, 0, 0, hInstance, NULL, NULL, NULL, NULL, "cedit" });

	menu = CreateMenu();
	submenu = CreatePopupMenu();
	AppendMenu(submenu, MF_STRING, ID_Open, "&Open...");
	AppendMenu(submenu, MF_STRING, ID_Save, "&Save...");
	AppendMenu(submenu, MF_SEPARATOR, 0, 0);
	AppendMenu(submenu, MF_STRING, ID_Exit, "E&xit");
	AppendMenu(menu, MF_STRING | MF_POPUP, (UINT)submenu, "&File");
	submenu = CreatePopupMenu();
	AppendMenu(submenu, MF_STRING, ID_Undo, "&Undo");
	AppendMenu(menu, MF_STRING | MF_POPUP, (UINT)submenu, "&Edit");
	AppendMenu(menu, MF_STRING, ID_About, "&About");
	
	mainwindow = CreateWindow(
		"main",
		"EarthBound Tile Editor",
		WS_OVERLAPPEDWINDOW | WS_MAXIMIZE,
		CW_USEDEFAULT, CW_USEDEFAULT,
		512 + 128 + 3 * x + GetSystemMetrics(SM_CXFRAME) * 2,
		256 + 131 + 2 * y + GetSystemMetrics(SM_CYFRAME) * 2 + GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYHSCROLL) + GetSystemMetrics(SM_CYMENU),
		NULL, menu, hInstance, NULL);

	SendMessage(mainwindow, WM_SETHOTKEY, HOTKEYF_CONTROL << 8 | 'Z', 0);

	tilewindow = CreateWindow("tiles", "Tiles", WS_CHILD | WS_CAPTION | WS_VISIBLE,
		0, 0, 512 + x, 256 + y, mainwindow, NULL, hInstance, NULL);
	arrangewindow = CreateWindow("arrange", "Arrangements", WS_CHILD | WS_CAPTION | WS_VISIBLE | WS_HSCROLL,
		0, 256 + y, 494 + x, 131 + y + GetSystemMetrics(SM_CYHSCROLL), mainwindow, NULL, hInstance, NULL);
	edit1window = CreateWindow("edit1", "Layer 1", WS_CHILD | WS_CAPTION | WS_VISIBLE,
		512 + x, 105, 64 + x, 64 + y, mainwindow, NULL, hInstance, NULL);
	edit2window = CreateWindow("edit2", "Layer 2", WS_CHILD | WS_CAPTION | WS_VISIBLE,
		512 + 64 + x + x, 105, 64 + x, 64 + y, mainwindow, NULL, hInstance, NULL);
	palwindow = CreateWindow("pal", "Palette", WS_CHILD | WS_CAPTION | WS_VISIBLE,
		512 + x, 105 + 64 + y, 128 + x, 32 + y, mainwindow, NULL, hInstance, NULL);
	peditwindow = CreateWindow("pedit", "Palette Edit", WS_CHILD | WS_CAPTION | WS_VISIBLE,
		512 + x, 105 + 64 + y + 32 + y, 129 + x, 48 + y, mainwindow, NULL, hInstance, NULL);
	aeditwindow = CreateWindow("aedit", "Arrange", WS_CHILD | WS_CAPTION | WS_VISIBLE,
		494 + x + 80 + x, 105 + 64 + y + 32 + y + 48 + y, 67 + x, 67 + y, mainwindow, NULL, hInstance, NULL);
	ceditwindow = CreateWindow("cedit", "Properties", WS_CHILD | WS_CAPTION | WS_VISIBLE,
		494 + x, 105 + 64 + y + 32 + y + 48 + y, 80 + x, 80 + y, mainwindow, NULL, hInstance, NULL);
	
	for(i = 0; i < 16; i++)
		CreateWindow("EDIT", NULL, WS_CHILD | WS_VISIBLE | WS_BORDER, 20 * (i % 4), 20 * (i / 4), 20, 20, ceditwindow, (HMENU)(ID_Prop + i), hInstance, NULL);

	SetScrollRange(arrangewindow, SB_HORZ, 0, 241, 0);

	listbox = CreateWindow(
		"COMBOBOX", NULL,
		WS_CHILD | WS_VISIBLE | WS_VSCROLL | CBS_DROPDOWNLIST,
		520, 20, 132, 150,
		mainwindow, (HMENU)ID_Selector, hInstance, NULL);
	for(i = 0; i < 20; i++)
		SendMessage(listbox, CB_ADDSTRING, 0, (LPARAM)tileset_name[i]);
	SendMessage(listbox, CB_SELECTSTRING, 0, (LPARAM)tileset_name[0]);

	CreateWindow(
		"COMBOBOX", NULL,
		WS_CHILD | WS_VISIBLE | WS_VSCROLL | CBS_DROPDOWNLIST,
		597, 50, 55, 150,
		mainwindow, (HMENU)ID_Palette, hInstance, NULL);

	listbox = CreateWindow(
		"COMBOBOX", NULL,
		WS_CHILD | WS_VISIBLE | WS_VSCROLL | CBS_DROPDOWNLIST,
		597, 80, 55, 150,
		mainwindow, (HMENU)ID_Minipal, hInstance, NULL);
	for(i = 0; i < 6; i++) {
		char buf[2];
		buf[0] = i + '0';
		buf[1] = 0;
		SendMessage(listbox, CB_ADDSTRING, 0, (LPARAM)buf);
	}
	SendMessage(listbox, CB_SELECTSTRING, 0, (LPARAM)"0");

	rscroll = CreateWindow("SCROLLBAR", NULL, WS_CHILD | WS_VISIBLE | SBS_HORZ,
		0, 0, 100, 16, peditwindow, NULL, hInstance, NULL);
	gscroll = CreateWindow("SCROLLBAR", NULL, WS_CHILD | WS_VISIBLE | SBS_HORZ,
		0, 16, 100, 16, peditwindow, NULL, hInstance, NULL);
	bscroll = CreateWindow("SCROLLBAR", NULL, WS_CHILD | WS_VISIBLE | SBS_HORZ,
		0, 32, 100, 16, peditwindow, NULL, hInstance, NULL);
	SetScrollRange(rscroll, SB_CTL, 0, 31, 0);
	SetScrollRange(gscroll, SB_CTL, 0, 31, 0);
	SetScrollRange(bscroll, SB_CTL, 0, 31, 0);

	ShowWindow(mainwindow, nCmdShow);
	UpdateWindow(mainwindow);

	while(GetMessage(&msg, NULL, 0, 0) > 0) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	return msg.wParam;
}
